/* ##########################################################################

    Copyright 2004-2009:
    ANTONIO JAVIER BARRAGAN, antonio.barragan@diesia.uhu.es
    http://www.uhu.es/antonio.barragan
    JOSE MANUEL ANDUJAR, andujar@diesia.uhu.es

    DPTO. DE ING. ELECTRONICA, DE SISTEMAS INFORMATICOS Y AUTOMATICA
    E.P.S. LA RABIDA, UNIVERSITY OF HUELVA (SPAIN)

    For more information, please contact with authors.

########################################################################## */

/*
    This software is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "jacobian.h"

int Jacobian::order() const {return n;}

double Jacobian::read_J(int row, int column) const
{
    if (row>n || column>n)
    {
        #ifndef MATLAB_MEX_FILE
            std::cout<<"Internal Error. Number of Inputs exceded in 'read_function'. Used "<<(row>column?row:column)<<", the maximum is "<<n-1<<std::endl;
        #endif
        #ifdef MATLAB_MEX_FILE
            mexPrintf("Internal Error. Number of Inputs exceded in 'read_function'. Used %d, the maximum is %d.\n",row>column?row:column,n-1);
        #endif
        return 0;
    }
    return J[row][column];
}

int Jacobian::write_J(double value, int row, int column)
{
    if (row>n || column>n)
    {
        #ifndef MATLAB_MEX_FILE
            std::cout<<"Internal Error. Number of Inputs exceded in 'read_function'. Used "<<(row>column?row:column)<<", the maximum is "<<(n-1)<<std::endl;
        #endif
        #ifdef MATLAB_MEX_FILE
            mexPrintf("Internal Error. Number of Inputs exceded in 'read_function'. Used %d, the maximum is %d.\n",row>column?row:column,n-1);
        #endif
        return 0;
    }
    J[row][column]=value;
    return 1;
}

int Jacobian::resolv(System S, System C, const double *const x)
{
    // Indices
    int i,k,q,j,r,s,l,p,v,M_max,N_max;

    // System and Controller constants
    // n = System order
    if (n!=S.outputs())
        return 0;
    if (S.outputs()!=C.inputs())
        return 0;
    if (n!=S.inputs()-C.outputs())
        return 0;
    const int m=C.outputs(); // Number of control inputs
    int *M=new int[n];
    M_max=0; // The maximum number of rules of the plant
    for (i=0;i<n;i++)
    {
        M[i]=S.rules(i); // Number of rules of the plant in the ith output
        if (M[i]>M_max)
            M_max=M[i];
    }
    int *N=new int[m];
    N_max=0; // The maximum number of rules of the controller
    for (j=0;j<m;j++)
    {
        N[j]=C.rules(j); // Number of rules of the plant in the jth output
        if (N[j]>N_max)
            N_max=N[j];
    }

    // Auxiliary values
    // using M_max and N_max to make the memory reservation
    Membership *P;
    Rule *R;

    double temp1=0,temp2=0,temp3=0,temp4=0;
    double Sum2=0,Sum3=0;
    double a1,a2,b1,b2,c1,c2;
    double at;  // Variable coefficient 'a' for the plant. 

    double *X=new double [n+1];
    X[0]=1;
    for (i=0;i<n;i++)
        X[i+1]=x[i];
    double *Sum_activ=new double[m];
    double *Sum_W=new double[n];
    double **activ=new double *[N_max]; // Matching degree of the rules of the controller (omega)
    double **ct=new double *[n+1]; // Variable coefficient for the controller
    double *u=new double [m]; // Control signal
    double **W=new double *[M_max]; // Matching degree of the rules of the plant
    double *bt=new double [m]; // Variable coefficient 'b' for the plant 
    double **der_u=new double *[m]; // du/dx
    double ***der_w=new double **[M_max]; // dw/dx
    double ***der_activ=new double **[N_max]; //domega/dx

    for (j=0;j<m;j++)
        der_u[j]=new double [n];
    for (l=0;l<M_max;l++)
    {
        W[l]=new double[n];
        der_w[l]=new double *[n];
        for (i=0;i<n;i++)
            der_w[l][i]=new double[n];
    }
    for (r=0;r<N_max;r++)
    {
        activ[r]=new double[m];
        der_activ[r]=new double *[n];
        for (j=0;j<m;j++)
            der_activ[r][j]=new double[n];
    }
    for (i=0;i<=n;i++)
        ct[i]=new double[m];

    // Calculation procedure
    for (j=0;j<m;j++)
    {
        Sum_activ[j]=0;
        for (r=0;r<N[j];r++)
        {
            R=C.read_Rule(r,j);
            activ[r][j]=1;
            for (k=0;k<n;k++)
            {
                P=R->read_function(k);
				if (P->type()!=NOMF) // NOMF doesn't affect to the system
					activ[r][j]*=P->eval(x[k]);
            }
            Sum_activ[j]+=activ[r][j];
        } // activ[r][j] is calculated here
        u[j]=0;
        for (k=0;k<=n;k++)
        {
            ct[k][j]=0;
            for (r=0;r<N[j];r++)
                ct[k][j]+=(activ[r][j]*C.read_Rule(r,j)->read_TSK(k));
            ct[k][j]=ct[k][j]/Sum_activ[j]; // ct[k][j] is calculated here
            u[j]+=(X[k]*ct[k][j]);
        } // u[j] is calculated here
    }
    for (i=0;i<n;i++)
    {
        Sum_W[i]=0;
        for (l=0;l<M[i];l++)
        {
            R=S.read_Rule(l,i);
            W[l][i]=1;
            for (k=0;k<n;k++)
            {
                P=R->read_function(k);
				if (P->type()!=NOMF) // NOMF doesn't affect to the system
					W[l][i]*=P->eval(x[k]);
            }
            for (j=0;j<m;j++)
            {
                P=R->read_function(j+n);
				if (P->type()!=NOMF) // NOMF doesn't affect to the system
					W[l][i]*=P->eval(u[j]);
            }
            // W[l][i] is calculated here
            Sum_W[i]+=W[l][i];
        }
    }
    for (j=0;j<m;j++)
    {
        for (q=0;q<n;q++)
        {
            for (r=0;r<N[j];r++)
            {
                R=C.read_Rule(r,j);
				P=R->read_function(q);
				if (P->type()!=NOMF) // NOMF doesn't affect to the system
				{
					der_activ[r][j][q]=P->evalder(x[q]);
					temp1=1;
					for (k=0;k<n;k++)
					{
						if (k!=q)
						{
							P=R->read_function(k);
							if (P->type()!=NOMF) // NOMF doesn't affect to the system
								temp1*=P->eval(x[k]);
						}
					}
					der_activ[r][j][q]*=temp1;
				}
				else
					der_activ[r][j][q]=0;
            }
            // der_activ[r][j][q] is calculated here
            temp1=0;
            for (k=0;k<=n;k++)
            {
                temp2=0;
                for (r=0;r<N[j];r++)
                {
                    c1=C.read_Rule(r,j)->read_TSK(k);
                    for (s=0;s<N[j];s++)
                    {
                        c2=C.read_Rule(s,j)->read_TSK(k);
                        temp2+=(der_activ[r][j][q]*activ[s][j]*(c1-c2));
                    }
                }
                temp2*=X[k];
                temp1+=temp2;
            }
            der_u[j][q]=ct[q+1][j]+temp1/pow(Sum_activ[j],2);
        } // der_u[j][q] is calculated here
    }
    for (i=0;i<n;i++)
    {
        for (j=0;j<m;j++)
        {
            bt[j]=0;
            for (l=0;l<M[i];l++)
                bt[j]=bt[j]+(W[l][i]*S.read_Rule(l,i)->read_TSK(j+n+1));
            bt[j]=bt[j]/Sum_W[i]; // bt[j][i] is calculated here
        }
        for (q=0;q<n;q++)
        {
            at=0;
            Sum2=0;
            for (j=0;j<m;j++)
                Sum2+=(bt[j]*ct[q+1][j]); // Sum2[i][q] is calculated here
            for (l=0;l<M[i];l++)
            {
                R=S.read_Rule(l,i);
                at+=(W[l][i]*R->read_TSK(q+1));
				P=R->read_function(q);
				if (P->type()!=NOMF) // NOMF doesn't affect to the system
				{
					temp1=P->evalder(x[q]);
					for(j=0;j<m;j++)
					{
						P=R->read_function(j+n);
						if (P->type()!=NOMF) // NOMF doesn't affect to the system
							temp1*=P->eval(u[j]);
					}
					temp3=1;
					for(k=0;k<n;k++)
					{
						if (k!=q)
						{
							P=R->read_function(k);
							if (P->type()!=NOMF) // NOMF doesn't affect to the system
								temp3*=P->eval(x[k]);
						}
					}
					temp1*=temp3;
					temp3*=R->read_function(q)->eval(x[q]); // R->read_function(q) != NOMF (see if)
					temp2=0;
					for (v=0;v<m;v++)
					{
						P=R->read_function(v+n);
						if (P->type()!=NOMF) // NOMF doesn't affect to the system
						{
							temp4=der_u[v][q]*P->evalder(u[v]);
							for (j=0;j<m;j++)
							{
								if (j!=v)
								{
									P=R->read_function(j+n);
									if (P->type()!=NOMF) // NOMF doesn't affect to the system
										temp4*=P->eval(u[j]);
								}
							}
							temp2+=temp4;
						} // else temp4=0 -> temp2 don't change						
					}
					der_w[l][i][q]=temp1+temp2*temp3;
				}
				else
					der_w[l][i][q]=0;
				// der_w[l][i][q] is calculated here
            }
            at/=Sum_W[i]; // at[q][i] is calculated here
            Sum3=0;
            for (k=0;k<=n;k++)
            {
                temp1=0;
                for (l=0;l<M[i];l++)
                {
                    a1=S.read_Rule(l,i)->read_TSK(k);
                    for (p=0;p<M[i];p++)
                    {
                        a2=S.read_Rule(p,i)->read_TSK(k);
                        temp1=temp1+der_w[l][i][q]*W[p][i]*(a1-a2);
                    }
                }
                temp1/=pow(Sum_W[i],2);
                temp2=0;
                for (j=0;j<m;j++)
                {
                    temp3=0;
                    for (l=0;l<M[i];l++)
                    {
                        b1=S.read_Rule(l,i)->read_TSK(j+n+1);
                        for (p=0;p<M[i];p++)
                        {
                            b2=S.read_Rule(p,i)->read_TSK(j+n+1);
                            temp3+=(der_w[l][i][q]*W[p][i]*(b1-b2));
                        }
                    }
                    temp3*=ct[k][j]/pow(Sum_W[i],2);
                    temp4=0;
                    for (r=0;r<N[j];r++)
                    {
                        c1=C.read_Rule(r,j)->read_TSK(k);
                        for (s=0;s<N[j];s++)
                        {
                            c2=C.read_Rule(s,j)->read_TSK(k);
                            temp4+=(der_activ[r][j][q]*activ[s][j]*(c1-c2));
                        }
                    }
                    temp4*=bt[j]/pow(Sum_activ[j],2);
                    temp2+=temp3+temp4;
                }
                Sum3+=((temp1+temp2)*X[k]);
            }
            J[i][q]=at+Sum2+Sum3; // J[i][q] is calculated here
        }
    }

    // Free dynamic memory
    for (j=0;j<m;j++)
        delete [] der_u[j];
    delete [] der_u;
    der_u=NULL;
    delete [] bt;
    bt=NULL;
    for (i=0;i<=n;i++)
        delete [] ct[i];
    delete [] ct;
    ct=NULL;
    for (l=0;l<M_max;l++)
    {
        delete [] W[l];
        for (i=0;i<n;i++)
            delete [] der_w[l][i];
        delete [] der_w[l];
    }
    delete [] W;
    W=NULL;
    delete [] der_w;
    der_w=NULL;
    for (r=0;r<N_max;r++)
    {
        delete [] activ[r];
        for (j=0;j<m;j++)
            delete [] der_activ[r][j];
        delete [] der_activ[r];
    }
    delete [] activ;
    activ=NULL;
    delete [] der_activ;
    der_activ=NULL;
    delete [] Sum_W;
    Sum_W=NULL;
    delete [] Sum_activ;
    Sum_activ=NULL;
    delete [] u;
    u=NULL;
    delete [] X;
    X=NULL;
    delete [] N;
    N=NULL;
    delete [] M;
    M=NULL;
    return 1;
}

// Use finite difference to aproximate the jacobian matrix. 'h' is the increment.
int Jacobian::aprox(System S, System C, const double * const x, double h)
{
	int i,j,k;
	double *x2, *dx, *dx2;

    // System and controller constants
    // n = System order
    if (n!=S.outputs())
        return 0;
    if (S.outputs()!=C.inputs())
        return 0;
    if (n!=S.inputs()-C.outputs())
        return 0;
    const int m=C.outputs(); // Number of control inputs

	dx=new double[n]; // Output of the plant in 'x'

	if (Evaluate(x,dx,S,C)!=1)
	{
		delete []dx;
		return 0;
	}

	x2=new double[n];
	dx2=new double[n];

	for (j=0;j<n;j++)
	{
		for (k=0;k<n;k++)
			x2[k]=x[k];
		x2[j]=x2[j]+h;

		if (Evaluate(x2,dx2,S,C)!=1) // Output of the plant in 'x2' (with 'h')
		{
			delete []x2;
			delete []dx;
			delete []dx2;
			return 0;
		}
		for (i=0;i<n;i++)
			J[i][j]=(dx2[i]-dx[i])/h; // Aproximation of the jacobian matrix
	}
	
	// Free dynamic memory
	delete []x2;
	delete []dx;
	delete []dx2;
	return 1;
}

Jacobian::Jacobian(int n)
{
    this->n=n;
    J=new double* [n];
    int row,column;
    for (row=0;row<n;row++)
    {
        J[row]=new double[n];
        for (column=0;column<n;column++)
                J[row][column]=0;
    }
}

Jacobian::Jacobian(System S, System C, const double * const x)
{
    int row,column;
    n=S.outputs();
    J=new double* [n];
    for (row=0;row<n;row++)
    {
        J[row]=new double[n];
        for (column=0;column<n;column++)
                J[row][column]=0;
    }
    resolv(S,C,x);
}

Jacobian::~Jacobian()
{
    int row;
    for (row=0;row<n;row++)
    {
        if (J[row]!=NULL)
        {
            delete [] J[row];
            J[row]=NULL;
        }
    }
    if (J!=NULL)
    {
        delete [] J;
        J=NULL;
    }
}
