/* ##########################################################################

    Copyright 2004-2009:
    ANTONIO JAVIER BARRAGAN, antonio.barragan@diesia.uhu.es
    http://www.uhu.es/antonio.barragan
    JOSE MANUEL ANDUJAR, andujar@diesia.uhu.es

    DPTO. DE ING. ELECTRONICA, DE SISTEMAS INFORMATICOS Y AUTOMATICA
    E.P.S. LA RABIDA, UNIVERSITY OF HUELVA (SPAIN)

    For more information, please contact with authors.

########################################################################## */

/*
    This software is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _MEMBERSHIP_H_
#define _MEMBERSHIP_H_

#include <iostream>
#include "ext_math.h"
#include "messages.h"

#if defined (MATLAB_MEX_FILE) // For compile in MATLAB
	#ifndef _MEX_H_ // In MATLAB are used 'mexPrintf' for generate outputs
		#define _MEX_H_
		#include "mex.h"
	#endif
#endif

#define MAX_SIZE_TYPE_NAME 16

/*
 TYPE_MF, MF_NAMES and MF_PARAM_NUMBER must be coherent.
 
 TYPE_MF is used to determinate the membership function type, MF_NAMES are
 the names of the membership functions, and MF_PARAM_NUMBER are the number of
 parameters of the membership functions.
 
 MAX_SIZE_TYPE_NAME is the maximum size of the name of the membership functions in MF_NAMES array.
 
 'createMF' must include all membership functions defined in this file
 
 IMPORTANT!!!: If the first or the last element of TYPE_MF is changed, is necesary check
 'membership.cpp' (others?) because this elements are used to make some comparisons.
*/

enum TYPE_MF // Implemented membership functions
{ // In brackets (number of parameters) = MF_PARAM_NUMBER
    NOMF=0, // This function make possible that a rule doesn't depend of a variable (0 params)
    BELLMF, // Bell (3 params)
	CONSTMF, // The Constant function always return the same value, its parameter (1 parameter)
    GAUSSMF, // Gaussian (2 params)
    GAUSS2MF, // Duoble gaussian (4 params)
    PIMF, // Pi function (4 params)
    PSIGMF, // Product of sigmoidals (4 params)
    SMF, // S (2 params)
    SIGMF, // Sigmoidal (2 params)
    SIG2MF, // Double sigmoidal (4 params)
    TRAPMF, // Trapeziodal (4 params)
    TRAP2MF, // Double trapezoidal (8 params)
    TRIMF, // Triangular (2 params)
    ZMF // Z function (2 params)
};

static const char *MF_NAMES[]=
{// Names of TYPE_MF Membership Functions - UPPERCASE
    "NOMF","BELLMF","CONSTMF","GAUSSMF","GAUSS2MF","PIMF","PSIGMF","SMF","SIGMF","SIG2MF","TRAPMF","TRAP2MF","TRIMF","ZMF"
};

static const int MF_PARAM_NUMBER[]=
{// Number of parameters of each membership function in TYPE_MF
    0, // NOMF
    3, // BELLMF
    1, // CONSTMF
    2, // GAUSSMF
    4, // GAUSS2MF
    4, // PIMF
    4, // PSIGMF
    2, // SMF
    2, // SIGMF
    4, // SIG2MF
    4, // TRAPMF
    8, // TRAP2MF
    3, // TRIMF
    2  // ZMF
};

// Declarations ################################################################

class Membership // -Parent class-
{
protected:
    double *param; // Vector of parameters that define the membership function
    int n;  // Number of parameters of the membership function
    TYPE_MF type_mf;
public:
    virtual double eval(double x) const;  // Evaluate the membership function
    virtual double evalder(double x) const; // Derived from the membership function
    int test(void) const; // If the parameters are erroneous, this function corrects them
    int n_param(void) const; // Read the number of parameters that define the membership function
    TYPE_MF type(void) const; // Return the membership function type
    int type(TYPE_MF type_mf); // Change the membership function type
    int edit(int indice,double valor); // Change a parameter. Return 0 if error
    double read(int indice) const;  // Read a parameter
    Membership &operator=(const Membership &P);
    Membership(const Membership &P);
    Membership();
    ~Membership();
};

Membership * createMF(TYPE_MF t); // Virtual constructor

// Nomf ########################################################################
class Nomf:public Membership // This membership function don't exist in MATLAB.
{ // This type is used to initializate rules
 public:
    double eval(double x) const;
    double evalder(double x) const;
    Nomf(void);
};

// Bellmf ######################################################################
class Bellmf:public Membership // In MATLAB is 'gbellmf'
{
 public:
    double eval(double x) const;
    double evalder(double x) const;
    int test(void) const; // Modify the parameters to be correct
    Bellmf(double a=0.25, double b=2.5, double c=0.5);
};

// Constmf ######################################################################
class Constmf:public Membership // This membership function don't exist in MATLAB by default. It is 'constmf.m'
{
 public:
    double eval(double x) const;
    double evalder(double x) const;
    Constmf(double c=1.0);
};

// Gaussmf #####################################################################
class Gaussmf:public Membership // In MATLAB is 'gaussmf'
{
 public:
    double eval(double x) const;
    double evalder(double x) const;
    int test(void) const; // If the parameters of the membership function are valid returns 0
    Gaussmf(double Center=0.5, double Beta=0.3); // Sigma_MATLAB=Beta/sqrt(2);
};

// Gauss2mf ####################################################################
class Gauss2mf:public Membership // In MATLAB is 'gauss2mf'
{
 public:
    double eval(double x) const;
    double evalder(double x) const;
    int test(void) const; // Modify the parameters to be correct
    Gauss2mf(double Center1=0.45, double Beta1=0.25, double Center2=0.55, double Beta2=0.25); // Sigma_MATLAB=Beta/sqrt(2);
};

// Pimf ########################################################################
class Pimf:public Membership // In MATLAB is 'pimf'
{
 public:
    double eval(double x) const;
    double evalder(double x) const;
    Pimf(double a=0.05, double b=0.45, double c=0.55, double d=0.95);
};

// PSigmf ################################################################
class PSigmf:public Membership // In MATLAB is 'psigmf'
{
 public:
    double eval(double x) const;
    double evalder(double x) const;
    PSigmf(double a1=11, double c1=0.25, double a2=-11, double c2=0.75);
};

// Smf #########################################################################
class Smf:public Membership // In MATLAB is 'smf'
{
 public:
    double eval(double x) const;
    double evalder(double x) const;
    Smf(double a=0.05, double b=0.45);
};

// Sigmf #################################################################
class Sigmf:public Membership // In MATLAB is 'sigmf'
{
 public:
    double eval(double x) const;
    double evalder(double x) const;
    Sigmf(double a=14, double c=0.25);
};

// Sig2mf ################################################################
class Sig2mf:public Membership // In MATLAB is 'dsigmf'
{
 public:
    double eval(double x) const;
    double evalder(double x) const;
    Sig2mf(double a1=11, double c1=0.25, double a2=11, double c2=0.75);
};

// Trapmf ###############################################################
class Trapmf:public Membership // In MATLAB is 'trapmf'
{
 public:
    double eval(double x) const;
    double evalder(double x) const;
    int test(void) const; // Modify the parameters to be correct
    Trapmf(double a=0.05, double b=0.45, double c=0.55, double d=0.95);
};

// Trap2mf #####################################################################
class Trap2mf:public Membership // This membership function don't exist in MATLAB by default. It is 'trap2mf.m'
{
 public:
    double eval(double x) const;
    double evalder(double x) const;
    int test(void) const; // Modify the parameters to be correct
    Trap2mf(double a1=0.0, double b1=0.2, double c1=0.3, double d1=0.4,double a2=0.6, double b2=0.8, double c2=0.9, double d2=1);
};

// Trimf ################################################################
class Trimf:public Membership // In MATLAB is 'trimf'
{
 public:
    double eval(double x) const;
    double evalder(double x) const;
    int test(void) const; // Modify the parameters to be correct
    Trimf(double a=0.0, double b=0.5, double c=1.0);
};

// Zmf #########################################################################
class Zmf:public Membership // In MATLAB is 'zmf'
{
 public:
    double eval(double x) const;
    double evalder(double x) const;
    Zmf(double a=0.55, double b=0.95);
};
#endif
